 /*
  * main.c
  * 
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
  * Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  * 
  * This program is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  * See the GNU General Public License for more details.
  * 
  * You should have received a copy of the GNU General Public License along
  * with this program.  If not, see <http://www.gnu.org/licenses/>.
  * 
  * See the COPYING file.
  */


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <syslog.h>
#include <unistd.h>
#include <signal.h>
#include <getopt.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <libgen.h>     // dirname() and basename()
#include <ctype.h>      // isdigit()	  
#include <errno.h>
#include <assert.h>
#include <stdbool.h>

#include <ubus/libubus.h>
#include <libubox/blobmsg_json.h>

#include "def.h"

#define MAX_SIZE 256

const char *progname="ulogout";

static const char *pid_file="/var/run/ulogout/ulogout.pid";
static bool foreground = false;
static int pid_fd = -1;  // lock file

static struct ubus_context *ctx;
static struct ubus_request_data req_data;
static struct blob_buf bb_t;

void parse_args(int argc, char *argv[]);
pid_t get_pid_t(void);
int wait_on_kill(int s, int m);
void daemonize(void);

enum {
	SHUTDOWN_ID,
	__SHUTDOWN_MAX,
};
 
static const struct blobmsg_policy shutdown_policy[] =
{
	[SHUTDOWN_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32},
};
 
enum {
	REBOOT_ID,
	__REBOOT_MAX,
};
 
static const struct blobmsg_policy reboot_policy[] =
{
	[REBOOT_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32},
};
 
enum {
	SUSPEND_ID,
	__SUSPEND_MAX,
};
 
static const struct blobmsg_policy suspend_policy[] =
{
	[SUSPEND_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32},
};
 
enum {
	HIBERNATE_ID,
	__HIBERNATE_MAX,
};
 
static const struct blobmsg_policy hibernate_policy[] =
{
	[HIBERNATE_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32},
};

// Usage: strncpy_t(str1, sizeof(str1), str2, strlen(str2));
// Copy string "in" with at most "insz" chars to buffer "out", which
// is "outsz" bytes long. The output is always 0-terminated. Unlike
// strncpy(), strncpy_t() does not zero fill remaining space in the
// output buffer:
static
char* strncpy_t(char* out, size_t outsz, const char* in, size_t insz)
{
   assert(outsz > 0);
   while(--outsz > 0 && insz > 0 && *in) { *out++ = *in++; insz--; }
   *out = 0;
   return out;
}

/**
 * \brief A subprogram like popen() but with diferences explained below
 * The pipe is to read in the current process and connected to stderr in the
 * new one. The intent is to report error messages. Plus pid is returned.
 * Disclaimer: The caller must close the pipe after use
 */
static 
FILE *epopen( const char *cmd,  pid_t *pid)
{
  int pipefd[2];
  pid_t p_id;
  FILE *pf;
  int rc;

  // Create the pipe
  if(pipe(pipefd)) {
	  
      *pid = -1;
      return NULL;
  }

  p_id = fork();
 
  switch(p_id) {
    
    case 0: /* child */
      close(pipefd[0]);
      close(STDERR_FILENO);
      dup2(pipefd[1], STDERR_FILENO);
      close(pipefd[1]);
      execl( "/bin/sh", "sh", "-c", cmd, (char *)NULL );
      fprintf( stderr, "%s: error invoking /bin/sh: %s\n", progname, strerror(errno) );
      exit(EXIT_FAILURE); /* reminder: we are the child */
    
    case -1: /* fork() failed ! */
      rc = errno;
      close(pipefd[0]);
      close(pipefd[1]);
      errno = rc;
      *pid = -1;
      return NULL;
    
    default: /* the parent */
      close(pipefd[1]);
      pf = fdopen(pipefd[0], "r");
      *pid = p_id;
      return pf;
  }
}

static int shutdown_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
    pid_t pid;
    FILE *pfin = NULL;
    int wstatus;
    char cmd[32]={0};
    int rc __attribute__((unused)) = 0;
	
	strncpy_t(cmd, sizeof(cmd), "/bin/sync", strlen("/bin/sync"));
    pfin = epopen(cmd, &pid);
    if(pfin) { 
       fclose(pfin);
       waitpid(pid, &wstatus, 0);
	}
	
	if(wstatus != 0) return -1;
	
	strncpy_t(cmd, sizeof(cmd), "/bin/chvt 1", strlen("/bin/chvt 1"));
    pfin = epopen(cmd, &pid);
    if(pfin) { 
       fclose(pfin);
       waitpid(pid, &wstatus, 0);
	}
	
	if(wstatus != 0) return -1;
	
	rc = system("/sbin/shutdown");

	blob_buf_init(&bb_t, 0);
	blobmsg_add_string(&bb_t,"Server reply - Request is being proceeded: ", "/sbin/shutdown");
	ubus_send_reply(ctx, req, bb_t.head);

	ubus_defer_request(ctx, req, &req_data);
	ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
	
	return 0;
}

static int reboot_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
    pid_t pid;
    FILE *pfin = NULL;
    int wstatus;
    char cmd[32]={0};
    int rc __attribute__((unused)) = 0;
	
	strncpy_t(cmd, sizeof(cmd), "/bin/sync", strlen("/bin/sync"));
    pfin = epopen(cmd, &pid);
    if(pfin) { 
       fclose(pfin);
       waitpid(pid, &wstatus, 0);
	}
	
	if(wstatus != 0) return -1;
	
	strncpy_t(cmd, sizeof(cmd), "/bin/chvt 1", strlen("/bin/chvt 1"));
    pfin = epopen(cmd, &pid);
    if(pfin) { 
       fclose(pfin);
       waitpid(pid, &wstatus, 0);
	}
	
	if(wstatus != 0) return -1;
	
	rc = system("/sbin/reboot");

	blob_buf_init(&bb_t, 0);
	blobmsg_add_string(&bb_t,"Server reply - Request is being proceeded: ", "/sbin/reboot");
	ubus_send_reply(ctx, req, bb_t.head);

	ubus_defer_request(ctx, req, &req_data);
	ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
	
	return 0;
}

static int suspend_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
    pid_t pid;
    FILE *pfin = NULL;
    int wstatus;
    char cmd[32]={0};
    int rc __attribute__((unused)) = 0;
	
	strncpy_t(cmd, sizeof(cmd), "/usr/sbin/pm-suspend", strlen("/usr/sbin/pm-suspend"));
    pfin = epopen(cmd, &pid);
    if(pfin) { 
       fclose(pfin);
       waitpid(pid, &wstatus, 0);
	}
	
	if(wstatus != 0) return -1;

	blob_buf_init(&bb_t, 0);
	blobmsg_add_string(&bb_t,"Server reply - Request is being proceeded: ", "/usr/sbin/pm-suspend");
	ubus_send_reply(ctx, req, bb_t.head);

	ubus_defer_request(ctx, req, &req_data);
	ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
	
	return 0;
}

static int hibernate_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
    pid_t pid;
    FILE *pfin = NULL;
    int wstatus;
    char cmd[32]={0};
    int rc __attribute__((unused)) = 0;
	
	strncpy_t(cmd, sizeof(cmd), "/usr/sbin/pm-hibernate", strlen("/usr/sbin/pm-hibernate"));
    pfin = epopen(cmd, &pid);
    if(pfin) { 
       fclose(pfin);
       waitpid(pid, &wstatus, 0);
	}
	
	if(wstatus != 0) return -1;

	blob_buf_init(&bb_t, 0);
	blobmsg_add_string(&bb_t,"Server reply - Request is being proceeded: ", "/usr/sbin/pm-hibernate");
	ubus_send_reply(ctx, req, bb_t.head);

	ubus_defer_request(ctx, req, &req_data);
	ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
	
	return 0;
}

static const struct ubus_method ulogout_methods[] =
{
	UBUS_METHOD("shutdown", shutdown_handler, shutdown_policy),
	UBUS_METHOD("reboot", reboot_handler, reboot_policy),
	UBUS_METHOD("suspend", suspend_handler, suspend_policy),
	UBUS_METHOD("hibernate", hibernate_handler, hibernate_policy),
};
 

static struct ubus_object_type ulogout_obj_type =
	UBUS_OBJECT_TYPE("ering.ulogout", ulogout_methods);
 

static struct ubus_object ulogout_obj=
{
	.name = "ering.ulogout", 
	.type = &ulogout_obj_type,
	.methods = ulogout_methods,
	.n_methods = ARRAY_SIZE(ulogout_methods),
	//.path=
};


int main(int argc, char **argv)
{      
   int ret;
   const char *ubus_socket="/var/run/ubus/ubus.sock";
	
   parse_args(argc, argv);
   
	if (foreground == 0) {
		
		daemonize();
	}

	openlog(argv[0], LOG_PID|LOG_CONS, LOG_DAEMON);
	syslog(LOG_INFO, "Started %s", progname);
	
	uloop_init();
 
	ctx=ubus_connect(ubus_socket);
	if(ctx==NULL) exit(EXIT_FAILURE);
 
	ubus_add_uloop(ctx);

	ret=ubus_add_object(ctx, &ulogout_obj);
	if(ret!=0) goto UBUS_FAIL;
	
	uloop_run();
	uloop_done();
	
UBUS_FAIL:
   ubus_free(ctx);

   if( access( pid_file, F_OK ) == 0 ) {
      if(pid_fd != -1) {
         int rc = lockf(pid_fd, F_ULOCK, 0);
         if(rc != 0) exit(EXIT_FAILURE);
         close(pid_fd);
      }        
      unlink(pid_file);
   }
   
   syslog(LOG_INFO, "Stopped %s", progname);
   closelog();
   
   return 0;
}

void parse_args(int argc, char *argv[])
{
   static struct option long_options[] = {
	   
      {"foreground",            no_argument, 0, 'f'},
      {"kill",                  no_argument, 0, 'k'},
      {"check-running",         no_argument, 0, 'c'},
      {0, 0, 0}
   };
   
   int value, option_index = 0;
   int _kill = 0, _check = 0;
   while ((value = getopt_long(argc, argv, "fkc", long_options, &option_index)) != -1) {
      switch (value) {
         case 0:
            if (long_options[option_index].flag)
            break;
            
         case 'f':
            foreground = true;
            break;
                
         case 'k':
            _kill = 1;
            break;
            
         case 'c':
            _check = 1;
            break;
                  
         default:
            syslog(LOG_ERR, "Unknown parameter.");
            break;
      }
      
   }
  
   if (_kill) {

      int rc = 0;
      
      if ((rc = wait_on_kill(SIGTERM, 60)) < 0) {
         log_error("Failed to kill daemon with SIGTERM: %s\n", strerror(errno));
         exit(1);
      }
        
      exit(EXIT_SUCCESS);
   }
   
   if (_check) {
      
      pid_t pid = get_pid_t();
      
       if (pid == (pid_t) -1 || pid == 0) {
          // Do not print anything here. Empty output is used in snetaid.postinst script 
          // to check whether or not sysvinit is supervising snetaid
          // printf("%s not running.\n", progname);
          exit(EXIT_FAILURE);
       } else {
          printf("%s process running as pid %u.\n", progname, pid);
          exit(EXIT_SUCCESS);
       }
   }
}
   
/**
 * \brief Create the directory to store the pidfile and the logfile.
 */
static
void mkdir_piddir(const char *s)
{
	char *piddir, *tmp;

	piddir = strdup(s);
	tmp = strrchr(piddir, '/');
	if (tmp) {
		*tmp = '\0';
		mkdir(piddir, 0755);
	}
	free(piddir);
}
   
/**
 * \brief Get the PID of the process.
 */
pid_t get_pid_t(void)
{
   static char txt[MAX_SIZE];
   int fd = -1;
   pid_t ret = (pid_t) -1, pid;
   ssize_t l;
   long lpid;
   char *e = NULL;
   
   if ((fd = open(pid_file, O_RDWR, 0644)) < 0) {
      if ((fd = open(pid_file, O_RDONLY, 0644)) < 0) {
         if (errno != ENOENT) {
         log_error("Failed to open pidfile '%s': %s , errno=%d\n", pid_file, strerror(errno), errno);
         }
         goto finish;
      }
   }
   
   if ((l = read(fd, txt, MAX_SIZE-1)) < 0) {
      int saved_errno = errno;
      log_error("read(): %s , errno=%d", strerror(errno), errno);
      unlink(pid_file);
      errno = saved_errno;
      goto finish;
   }
   
   txt[l] = 0;
   txt[strcspn(txt, "\r\n")] = 0;
   
   errno = 0;
   lpid = strtol(txt, &e, 10);
   pid = (pid_t) lpid;
   
   if (errno != 0 || !e || *e || (long) pid != lpid) {
      unlink(pid_file);
      errno = EINVAL;
      log_warn("PID file corrupt, removing. (%s) , errno=%d", pid_file, errno);
      goto finish;
   }
   
   if (kill(pid, 0) != 0 && errno != EPERM) {
      int saved_errno = errno;
      log_warn("Process %lu died: %s; trying to remove PID file. (%s)", (unsigned long) pid, strerror(errno), pid_file);
      unlink(pid_file);
      errno = saved_errno;
      goto finish;
   }
   
   ret = pid;

finish:
   if (fd != -1) {
      int rc = 0;
      if((rc=lockf(fd, F_ULOCK, 0)) < 0) {
		  log_error("%s: Cannot unlock file\n", progname);
		  exit(EXIT_FAILURE);
	  }
	  close(fd);
   }
   
   return ret;
}

/**
 * \brief Kill the daemon process.
 */
int wait_on_kill(int s, int m)
{
   pid_t pid;
   time_t t;

   if ((pid = get_pid_t()) < 0) 
      return -1;

   if (kill(pid, s) < 0)
      return -1;

   t = time(NULL) + m;

   for (;;) {
      int r;
      struct timeval tv = { 0, 100000 };

      if (time(NULL) > t) {
         errno = ETIME;
         return -1;
      }

      if ((r = kill(pid, 0)) < 0 && errno != ESRCH)
         return -1;

      if (r)
         return 0;

      if (select(0, NULL, NULL, NULL, &tv) < 0)
         return -1;
    }
}

/**
 * \brief This function will daemonize this app
 */
void daemonize()
{
   int ret;
   pid_t pid = 0;
   int fd;

   /* Fork off the parent process.
    * The first fork will change our pid
    * but the sid and pgid will be the
    * calling process
    */
   pid = fork();

   /* An error occurred */
   if (pid < 0) {
      exit(EXIT_FAILURE);
   }

   /* Fork off for the second time.
    * The magical double fork. We're the session
    * leader from the code above. Since only the
    * session leader can take control of a tty
    * we will fork and exit the session leader.
    * Once the fork is done below and we use
    * the child process we will ensure we're
    * not the session leader, thus, we cannot
    * take control of a tty.
    */
   if (pid > 0) {
      exit(EXIT_SUCCESS);
   }

   /* On success: The child process becomes session leader */
   if (setsid() < 0) {
      exit(EXIT_FAILURE);
   }

   /* Ignore signal sent from child to parent process */
   signal(SIGCHLD, SIG_IGN);

   /* Fork off for the second time*/
   pid = fork();

   /* An error occurred */
   if (pid < 0) {
      exit(EXIT_FAILURE);
   }

   /* Success: Let the parent terminate */
   if (pid > 0) {
      exit(EXIT_SUCCESS);
   }
   
   /* Set new file permissions */
   umask(0);

   /* Change the working directory to /tmp */
   /* or another appropriated directory */
   ret = chdir("/tmp");
   if(ret != 0) exit(EXIT_FAILURE);

   /* Close all open file descriptors */
   for (fd = sysconf(_SC_OPEN_MAX); fd > 0; fd--) {
      close(fd);
   }

   /* Reopen stdin (fd = 0), stdout (fd = 1), stderr (fd = 2) */
   stdin  = fopen("/dev/null", "r");
   stdout = fopen("/dev/null", "w+");
   stderr = fopen("/dev/null", "w+");
   
   /* Try to write PID of daemon to lockfile */
   if (pid_file)
   {
	  mkdir_piddir(pid_file);
	  
      pid_fd = open(pid_file, O_RDWR|O_CREAT, 0640);
      if (pid_fd < 0) { 
		 log_error("%s: Cannot open pidfile: %s\n", progname, strerror(errno));
         exit(EXIT_FAILURE);
      }
      
      if (lockf(pid_fd, F_TLOCK, 0) < 0) {
         /* Can't lock file */
         exit(EXIT_FAILURE);
      }

      // Get current PID
      char *mypid = NULL;
      if(asprintf(&mypid, "%jd", (intmax_t) getpid()) != -1) {
         ssize_t sz = 0;
         // Write PID to lockfile
         sz = write(pid_fd, mypid, strlen(mypid));
         if(sz == -1) { 
		    log_error("%s: Unable to write pid: %s\n", progname, strerror(errno));
			exit(EXIT_FAILURE);
	     }
	  } else {  
		 log_error("%s: Unable to write pid: %s\n", progname, strerror(errno));
		 exit(EXIT_FAILURE);
	  }
   }
}
